# -*- coding: utf-8 -*-
"""
Created on Wed Mar 31 09:58:03 2021

@author: yanni
"""
                    
import snappy  
from snappy import GPF 
from snappy import ProductIO
from snappy import jpy
import time 
import datetime
import os
import gdal
import rasterio
from rasterio.plot import show
from matplotlib import pyplot
import zipfile

# adjust footprint 
path = r'F:/Daten/Dokumente/Uni/Master/Wintersemester 20 21/Remote Sensing of the Cryosphere/UI/data/'
outpath = r'F:/Daten/Dokumente/Uni/Master/Wintersemester 20 21/Remote Sensing of the Cryosphere/UI/output/'
ndvi_path = r'F:/Daten/Dokumente/Uni/Master/Wintersemester 20 21/Remote Sensing of the Cryosphere/UI/ndvi/'
ndsi_path = r'F:/Daten/Dokumente/Uni/Master/Wintersemester 20 21/Remote Sensing of the Cryosphere/UI/ndsi/'

for zip_file in os.listdir(path):
    if zip_file.endswith('.zip'):
        with zipfile.ZipFile(path + zip_file, 'r') as zip_ref:
            zip_ref.extractall(path)
        os.remove(path + zip_file)
        
jpy = snappy.jpy
BandDescriptor = jpy.get_type('org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor')

def do_resampling(source):
    # Resamples all the bands to 30x30 m (B3)
    print('Resampling')
    parameters = snappy.HashMap()
    parameters.put('Resample', True)
    parameters.put('referenceBand', 'B3') # basically we use B3
    parameters.put('upsampling', 'Nearest')
    parameters.put('downsampling', 'First')
    parameters.put('flagDownsampling', 'First')
    output = snappy.GPF.createProduct('Resample', parameters, source)
    return output

def do_water_Mask(source):
    parameters = snappy.HashMap()
    water_mask = BandDescriptor()
    water_mask.name = 'Water_Mask'
    water_mask.type = 'float32'
    water_mask.expression = ('scl_water')
    targetBand_water = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetBand_water[0] = water_mask
    parameters.put('targetBands', targetBand_water)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output 

def cloud_mask_basic(source):
    parameters = snappy.HashMap()
    cm_basic = BandDescriptor()
    cm_basic.name = 'Cloud_Mask'
    cm_basic.type = 'float32'
    cm_basic.expression = ('quality_cloud_confidence')
    targetBand_cloudmask = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetBand_cloudmask[0] = cm_basic
    parameters.put('targetBands', targetBand_cloudmask)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output
    
def do_subset(source):
    # Defines the research area over Zugspitze 
    print('Subset')
    parameters = snappy.HashMap()
    parameters.put('copyMetadata', True)
    # adjust footprint
    footprint = "POLYGON((13.278350830078125 52.453917235802294,13.278350830078125 52.586780621860804,13.521423339843752 52.586780621860804,13.521423339843752 52.453917235802294,13.278350830078125 52.453917235802294))"
    parameters.put('geoRegion', footprint)
    sBands = "B1,B2,B3,B4,B8,B11,B12,Cloud_Mask,Water_Mask"
    parameters.put('sourceBands', sBands )
    output = snappy.GPF.createProduct('Subset', parameters, source)    
    return output

def do_BandMerge(master, slave):
    # merges the last two products 
    print('Band Merge')
    sourceProducts= snappy.HashMap()
    sourceProducts.put('masterProduct', master)
    sourceProducts.put('slaveProduct', slave)
    parameters = snappy.HashMap()
    output = snappy.GPF.createProduct('Merge', parameters, sourceProducts)
    return output

def do_CloudMask(source):
    # Creates a Cloud Mask and the NDSI
    print('Cloud Mask')
    parameters = snappy.HashMap()
    cloud_mask = BandDescriptor()
    cloud_mask.name = 'CloudMask'
    cloud_mask.type = 'float32'
    cloud_mask.expression = ('if Cloud_Mask > 90 then 2 else if Cloud_Mask > 50 AND B8 > 0.3 then 1 else 0')
    targetband_cloud = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',1)
    targetband_cloud[0] = cloud_mask
    parameters.put('targetBands', targetband_cloud)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output
    
def do_NDSI(source):
    #NDSI
    print('NDSI')
    parameters = snappy.HashMap()
    ndsi = BandDescriptor()
    ndsi.name = 'NDSI'
    ndsi.type = 'float32'
    ndsi.expression = ('(B3-B11)/(B3+B11)')
    
    targetband_ndsi = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',1)
    
    targetband_ndsi[0] = ndsi
    parameters.put('targetBands', targetband_ndsi)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output

def do_NDVI(source):
    print('NDVI')
    # Creates the Normalized Difference Vegetation Index
    parameters = snappy.HashMap()
    ndvi = BandDescriptor()
    ndvi.name= 'NDVI'
    ndvi.type = 'float32'
    ndvi.expression = ('(B8-B4)/(B8+B4)')
    targetband_ndvi = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetband_ndvi[0] = ndvi
    parameters.put('targetBands', targetband_ndvi)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output
    
def do_snow_pass_1(source):
    print('Snow Pass 1')
    # Picks up snow pixels, except in shadow regions
    parameters = snappy.HashMap()
    snow_pass1 = BandDescriptor()
    snow_pass1.name = 'snow_pass_1'
    snow_pass1.type = 'float32'
    snow_pass1.expression = ('if CloudMask < 2 AND NDSI > 0.4 AND B4 > 0.2 AND NDVI < 0 AND Water_Mask == 0 then 1 else 0')
        
    targetband_sp1 = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetband_sp1[0] = snow_pass1
    parameters.put('targetBands', targetband_sp1)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output

def do_pre_product_snow_cover(source):
    # Picks up the snow pixels in shadow regions with NDSI > 0.15 AND B4 > 0.04
    print('Pre-Product')
    parameters = snappy.HashMap()
    snow_mask = BandDescriptor()
    snow_mask.name = 'snowmask'
    snow_mask.type = 'float32'
    snow_mask.expression = ('if CloudMask < 2 AND NDSI > 0.150 AND B4 > 0.04 AND NDVI < 0 AND Water_Mask == 0 OR snow_pass_1 == 1 then 1 else if CloudMask > 0 then 9 else 0')
        
    targetband_ppsc = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetband_ppsc[0] = snow_mask
    parameters.put('targetBands', targetband_ppsc)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output

def do_snow_cover(source):
    # Binary TIFF with 1 == snow and 0 == no snow
    print('Snow Cover Result')
    parameters = snappy.HashMap()
    snow_mask2 = BandDescriptor()
    snow_mask2.name = 'snow_cover'
    snow_mask2.type = 'float32'
    snow_mask2.expression = ('if snowmask == 1 then 1 else 0')
        
    targetband_sc = jpy.array(
            'org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor',
            1
            )
    targetband_sc[0] = snow_mask2
    parameters.put('targetBands', targetband_sc)
    output = snappy.GPF.createProduct('BandMaths', parameters, source)
    return output

def main():
    
    if not os.path.exists(outpath):
        os.makedirs(outpath)
        
    if not os.path.exists(ndvi_path):
        os.makedirs(ndvi_path)
        
    if not os.path.exists(ndsi_path):
        os.makedirs(ndsi_path)
        
    
    for files in os.listdir(path):
        snowcover_filepath = outpath + files[0:60] + '_snowcover.tif'
        
        if os.path.exists(snowcover_filepath):
            print("snow cover output file already exists")
        else:
            loopstarttime=str(datetime.datetime.now())
            print('Start time:', loopstarttime)
            start_time = time.time()
            print('reading file...')
            sentinel_2 = snappy.ProductIO.readProduct(path + files)
            Resampling = do_resampling(sentinel_2)
            WaterMask = do_water_Mask(Resampling)
            BandMerge = do_BandMerge(Resampling, WaterMask)
            del Resampling
            del WaterMask
            CloudMask_Basic = cloud_mask_basic(BandMerge)
            BandMerge999 = do_BandMerge(BandMerge, CloudMask_Basic)
            Subset = do_subset(BandMerge999)
            del BandMerge
            CloudMask = do_CloudMask(Subset)
            BandMerge1 = do_BandMerge(Subset, CloudMask)
            del Subset
            del CloudMask
            NDSI = do_NDSI(BandMerge1)
            snappy.ProductIO.writeProduct(NDSI, ndsi_path + files[0:60] + '_ndsi', 'GeoTIFF')
            BandMerge2 = do_BandMerge(BandMerge1, NDSI)
            del NDSI
            del BandMerge1
            NDVI = do_NDVI(BandMerge2)
            snappy.ProductIO.writeProduct(NDVI, ndvi_path + files[0:60] + '_ndvi', 'GeoTIFF')
            BandMerge3 = do_BandMerge(BandMerge2, NDVI)
            del NDVI
            del BandMerge2
            SnowPass1 = do_snow_pass_1(BandMerge3)
            BandMerge4 = do_BandMerge(BandMerge3, SnowPass1)
            del SnowPass1
            del BandMerge3
            Snow_Cover_pre_product = do_pre_product_snow_cover(BandMerge4)
            BandMerge5 = do_BandMerge(BandMerge4, Snow_Cover_pre_product)
            del Snow_Cover_pre_product
            del BandMerge4
            SnowCover = do_snow_cover(BandMerge5)
            print("********************************")
            print(SnowCover)
            del BandMerge5
            print("writing output file...")
            snappy.ProductIO.writeProduct(SnowCover, outpath + files[0:60] + '_snowcover', 'GeoTIFF')
            sentinel_2.dispose()
            sentinel_2.closeIO()  
            del SnowCover 
            
            print("--- %s seconds ---" % (time.time() - start_time))
                  
    for filename in os.listdir(outpath):
        if filename.endswith('.tif'):
            filepath = os.path.join(outpath,filename)
            ds = gdal.Open(filepath)
            stats = ds.GetRasterBand(1).GetStatistics(1,1)
            snowcover_percentage = stats[2]*100
            print(snowcover_percentage)
            
    for file in os.listdir(ndsi_path): 
        day = file[17:19]
        month = file[15:17]
        year = file[11:15]
        filepath = ndsi_path + file
        src = rasterio.open(filepath)
        title_ndsi = ('NDSI for %s - %s - %s' %(day,month,year))
        show(src.read(1), transform=src.transform, cmap='gray', title=title_ndsi)
        pyplot.show()
        
    for file in os.listdir(ndvi_path):
        day = file[17:19]
        month = file[15:17]
        year = file[11:15]
        filepath = ndvi_path + file
        src = rasterio.open(filepath)
        title_ndvi = ('NDVI for %s - %s - %s' %(day,month,year))
        show(src.read(1), transform=src.transform, cmap='Greens', title=title_ndvi)
        pyplot.show()
      
    print('done')
    
if __name__== "__main__":
    main()       